from igraph import *
from time import *
import time
import sys


# computes the relevant and irrelevant vertices of a graph
# vertex u is irrelevant if there exists a vertex w such that N(u) \subseteq N[w]
# second parameter are open neighborhoods of the graph
# third parameter is number of irrelevant vertices
# fourth parameter is time to compute app irrelevant vertices
# last return parameter is a boolean (1 if computation finished within the time limit and 0 otherwise)

def neighborhood_relation(graph,  time_max):
    time_before_irr = time_ns()
    # vertices which will be deleted from the graph
    #deleted = set()
    # original indices
    n = graph.vcount()
    irr = [0 for i in range(n)]
    n3 = n**3
    indices = range(n)
    indices2 = range(n)
    deg = graph.degree()
    # array with neighborhoods
    neighborhoods = dict()
    open_neighborhoods = dict()
    # value and values will be our hashes for closed and open neighborhood
    # let v\in V: value(v)=|N[v]| + |N[N[v]]|*n + sum_v in N[v](id(v))
    values = []
    values2 = []
    value_dict = dict()
    for index in indices:
        val = deg[index]
        degs = 0
        ids = 0
        for vertex in graph.neighbors(index):
            degs += deg[vertex]
            ids += vertex
        val += n*degs + ids*n3
        values.append(val + index*n3)
        values2.append(val)
        value_dict[index]=val
        neighborhoods[index] = (set(graph.neighbors(index)).union({index}))
        open_neighborhoods[index] = (set(graph.neighbors(index)))
    # sort the vertices according to their hash value
    sizes = sorted(list(zip(values, indices)), reverse=True)
    values, indices = list(zip(*sizes))

    # If two vertices are in the same critical clique, they have the same value
    # in the following check for two subsequent vertices with the same value if they in fact have the same closed neighborhood
    index = 0
    neighbors = [set(graph.neighbors(indices[0])).union({indices[0]})]
    value = values[0]
    # the maximum size of a critical clique is k, all remaining vertices in this critical clique can be removed
    size = [1]
    unset_vertices = [indices[0]]
    used = set()
    numb_crit_cliques = 0
    crit_indices = [0]
    for i in range(1, n):
        val_i = values[i]
        vertex = indices[i]
        if value == val_i:
            # determine neighborhood of this vertex
            new_neigs = neighborhoods[vertex]#set(graph.neighbors(vertex)).union({vertex})
            # test if this new neighborhood was already observed
            flag = False
            lint = 0
            for neig in neighbors:
                if neig == new_neigs:
                    if size[lint] >= 1:
                        irr[vertex] = 1
                        #graph.vs[vertex]["irr"] = 1
                        #deleted.add(vertex)
                    else:
                        #if size[lint] == 1:
                        #    crit_indices[lint] = numb_crit_cliques
                        #    graph.vs[unset_vertices[lint]]["crit"] = index + numb_crit_cliques
                        #    used.add(unset_vertices[lint])
                        #    numb_crit_cliques += 1
                        used.add(vertex)
                    size[lint] += 1
                    flag = True
                    break
                lint += 1
            if flag == False:
                neighbors.append(new_neigs)
                size.append(1)
                unset_vertices.append(vertex)
                crit_indices.append(0)
        else:
            neighbors = [set(graph.neighbors(vertex)).union({vertex})]
            value = val_i
            index += numb_crit_cliques
            numb_crit_cliques = 0
            size = [1]
            unset_vertices = [vertex]
            crit_indices = [0]

    sizes = sorted(zip(values2, indices2), reverse=True)
    values2, indices = list(zip(*sizes))
    #find first vertex that is not a critical clique
    i = 0
    while irr[indices[i]] == 1 or indices[i] in used:
    #while graph.vs[indices[i]]["irr"] == 1 or indices[i] in used:
        i += 1
    index += numb_crit_cliques
    neighbors = [set(graph.neighbors(indices[i]))]
    value = values2[i]
    size = [1]
    for i in range(i+1, n):
        vertex = indices[i]
        if irr[vertex] == 1 or vertex in used:
        #if graph.vs[vertex]["irr"] == 1 or vertex in used:
            continue
        val_i = values2[i]
        if value == val_i:
            # determine neighborhood of this vertex
            new_neigs = open_neighborhoods[vertex]
            flag = False
            lint = 0
            for neig in neighbors:
                if neig == new_neigs:
                    flag = True
                    if size[lint] >= 1:
                        irr[vertex] = 1
                        #graph.vs[vertex]["irr"] = 1
                        #deleted.add(vertex)
                    #else:
                    #    graph.vs[vertex]["crit"] = index + lint
                    #    used.add(vertex)
                    #    size[lint] += 1
                    break
                lint += 1
            if flag == False:
                neighbors.append(new_neigs)
                size.append(1)
        else:
            neighbors = [open_neighborhoods[vertex]]
            value = val_i
            index += size.__len__()
            size = [1]

    # deletion of twin vertices is done at end, after dom dag is computed

    # now: Dom dag stuff
    indices = range(n)
    deg = graph.degree()
    sizes = sorted(list(zip(deg, indices)), reverse=True)
    unzipped = list(zip(*sizes))
    indices = unzipped[1]
    #index += size.__len__()
    root_set = set()
    print("Computed twin sets in", time.process_time(), "seconds.")

    for i in range(n):
        # check running time after each vertex
        if time.process_time() > time_max:
            numb_relevant = 0
            for i in range(n):
                if irr[i] == 0:
                #if graph.vs[i]["irr"] is None:
                    #graph.vs[i]["irr"] = 0
                    numb_relevant += 1
            numb_relevant = graph.vcount() - numb_relevant
            time_irr = (time_ns() - time_before_irr) / 1000000000
            return graph, open_neighborhoods, numb_relevant, time_irr, irr, False
        vertex = indices[i]
        if irr[vertex] == 1:
        #if graph.vs[vertex]["irr"] == 1:
            continue
        new = []
        if graph.degree(vertex) != 0:
            x = graph.neighbors(vertex)[0]
            for u in graph.neighbors(vertex):
                if graph.degree(u) < graph.degree(x):
                    x = u
            root_cand = set(neighborhoods[x])
            root_cand.remove(vertex)
            root_cand = root_cand.intersection(root_set)
            for u in graph.neighbors(vertex):
                root_cand = root_cand.intersection(neighborhoods[u])
                if len(root_cand) == 0:
                    break
            for r in root_cand:
                new.append(r)
        # if new is still empty, create a new root
        if not new:
            root_set.add(vertex)
            # do not do all these checks again for each vertex in the same critical clique
            #for ver in cliques[graph.vs[vertex]["crit"]]:
            #    deleted.add(ver)
        # this vertex is irrelevant
        else:
            irr[vertex] = 1
            #graph.vs[vertex]["irr"] = 1
            #deleted.add(vertex)
            #uused = []
            #while new != []:
            #    this_node = new.pop()
            #
            #    # check all direct children of this_node, if there are unvisited vertices and check if neighborhood of vertex is still a subset
            #    # if yes, add these vertices to the stack and if not add this vertex to the children array
            #    flag = False
            #    for sub_vertex in children[this_node]:
            #        # necessary condition: smaller degree and not larger hash
            #        if graph.degree(vertex) < graph.degree(sub_vertex) and value_dict[vertex] <= value_dict[sub_vertex]:
            #            if open_neighborhoods[vertex] < neighborhoods[sub_vertex]:
            #                if visited[sub_vertex] == False:
            #                    new.append(sub_vertex)
            #                visited[sub_vertex] = True
            #                uused.append(sub_vertex)
            #                flag = True
            #    if flag == False:
            #        children[this_node].append(vertex)
            #    # add structure for all vertices in that critical clique
            #    # append for every vertex in the critical clique of this_node
            #    for crit_node in cliques[graph.vs[this_node]["crit"]]:
            #        for ver in cliques[graph.vs[vertex]["crit"]]:
            #            structure[crit_node].append(ver)
            ## now forbid all vertices of that critical clique
            #for ver in cliques[graph.vs[vertex]["crit"]]:
            #    new_deleted.add(ver)
            ## undo uused
            #for entry in uused:
            #    visited[entry] = False

    # finally set the attribute of each relevant vertex to 0
    numb_relevant = 0
    for i in range(n):
        if irr[i] == 0:
        #if graph.vs[i]["irr"] is None:
            #graph.vs[i]["irr"] = 0
            numb_relevant += 1
    numb_relevant = graph.vcount() - numb_relevant
    #graph.delete_vertices(deleted)
    print("Number of relevant vertices: " + str(numb_relevant))
    print("Computed irrelevant vertices in", time.process_time(), "seconds.")

    # for i in range(graph.vcount()):
    #     if graph.vs[i]["irr"] == 1:
    #         continue
    #     for j in range(graph.vcount()):
    #         if j <= i or graph.vs[j]["irr"] == 1:
    #             continue
    #         neigs = open_neighborhoods[j]
    #         flag = False
    #         for neig in neigs:
    #             if neig not in neighborhoods[i]:
    #                 flag = True
    #         if flag == False:
    #             print(i,j)

    time_irr = (time_ns() - time_before_irr) / 1000000000
    return graph, open_neighborhoods, numb_relevant, time_irr, irr, True


